use std::fmt;

/// Error returned by `TryFrom` implementations between `Scalar`, `Array` and `Hash`.
#[derive(Debug)]
pub struct CastError;

impl std::error::Error for CastError {}

impl fmt::Display for CastError {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.write_str("wrong type")
    }
}

/// Generic errors from the perlmod crate.
#[derive(Clone, Debug)]
pub struct Error(pub(crate) String);

impl std::error::Error for Error {}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "error: {}", self.0)
    }
}

impl Error {
    #[inline]
    pub fn new(s: &str) -> Self {
        Self(s.to_string())
    }

    #[inline]
    pub fn new_owned(s: String) -> Self {
        Self(s)
    }

    #[inline]
    pub fn fail<T>(s: &str) -> Result<T, Self> {
        Err(Self(s.to_string()))
    }
}

impl serde::de::Error for Error {
    fn custom<T: std::fmt::Display>(msg: T) -> Self {
        Self(msg.to_string())
    }
}

impl serde::ser::Error for Error {
    fn custom<T: std::fmt::Display>(msg: T) -> Self {
        Self(msg.to_string())
    }
}

/// Error used in the `TryFrom` implementation generated by the
/// [`declare_magic!`](crate::declare_magic!) macro.
#[derive(Debug)]
pub enum MagicError {
    /// The value we tried to dereference was not a perl reference type.
    NotAReference(&'static str),

    /// The expected magic type was not found.
    NotFound(&'static str),
}

impl fmt::Display for MagicError {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match self {
            MagicError::NotAReference(class) => {
                write!(f, "value blessed into {} was not a reference", class)
            }
            MagicError::NotFound("") => {
                write!(f, "value did not contain the requested magic pointer")
            }
            MagicError::NotFound(class) => write!(
                f,
                "value blessed into {} did not contain its declared magic pointer",
                class
            ),
        }
    }
}

impl std::error::Error for MagicError {}
